import { Request, Response } from "express";
import Newsletter from "../models/newsletter";
import { newsletterLogger } from "../logger/logger";
import { newNewsletterId } from "../utils/utils";
import { isEmailValid } from "../validator/validator";
import rabbitmq from "../rabbitmq/rabbitmq";

const NEWSLETTER_STATE = {
  NOT_CONFIRMED: 0,
  CONFIRMED: 1,
};

export async function AddNewsletterParticipant(req: Request, res: Response) {
  try {
    const { type, email } = req.body;

    if (
      !type ||
      !email ||
      !(await isEmailValid(email, false)) ||
      !process.env.NEWSLETTER_ALLOWED_TYPES?.split(",").includes(type)
    ) {
      newsletterLogger.warn(
        "AddNewsletterParticipant: invalid request - missing email or type"
      );
      return res.status(400).send({ err: "invalid request" });
    }

    // only add the participant if they are not already in the newsletter

    const existingNewsletterParticipant = await Newsletter.findOne({
      where: {
        type: type,
        email: email,
      },
    });

    if (!existingNewsletterParticipant) {
      const newsletterId = newNewsletterId();

      await Newsletter.create({
        newsletter_id: newsletterId,
        email: email,
        type: type,
        state: NEWSLETTER_STATE.NOT_CONFIRMED,
      });

      newsletterLogger.info("AddNewsletterParticipant", email);

      rabbitmq.sendEmail(email, "confirmNewsletterParticipation", "de", {
        confirmNewsletterParticipationLink: `${process.env.NEWSLETTER_WEBSITE_URL}confirm?n=${newsletterId}`,
        unsubscribeNewsletterParticipationLink: `${process.env.NEWSLETTER_WEBSITE_URL}unsubscribe?n=${newsletterId}`,
        privacyPolicyLink:
          process.env.NEWSLETTER_PRIVACY_POLICY_URLS?.split(",")[0],
      });
    } else {
      newsletterLogger.warn(
        `AddNewsletterParticipant: email: ${email} already in newsletter of type: ${type}`
      );
    }

    return res.status(200).send({ status: "success" });
  } catch (error) {
    newsletterLogger.error("AddNewsletterParticipant err:", error as string);
    res.status(500).send({ err: "invalid request" });
  }
}

export async function ConfirmNewsletterParticipation(
  req: Request,
  res: Response
) {
  try {
    const { newsletterId } = req.params;

    if (!newsletterId) {
      newsletterLogger.warn(
        "ConfirmNewsletterParticipation: invalid request - missing email or type"
      );
      return res.status(400).send({ err: "invalid request" });
    }

    const newsletterParticipant = await Newsletter.findOne({
      where: {
        newsletter_id: newsletterId,
      },
    });

    if (!newsletterParticipant) {
      newsletterLogger.warn(
        `ConfirmNewsletterParticipation: newsletter id: ${newsletterId} not found`
      );
      return res.status(400).send({ err: "invalid request" });
    }

    if (newsletterParticipant.state !== NEWSLETTER_STATE.CONFIRMED) {
      newsletterParticipant.state = NEWSLETTER_STATE.CONFIRMED;
      await newsletterParticipant.save();

      rabbitmq.sendEmail(
        newsletterParticipant.email,
        "confirmedNewsletterParticipation",
        "de",
        {
          unsubscribeNewsletterParticipationLink: `${process.env.NEWSLETTER_WEBSITE_URL}unsubscribe?n=${newsletterId}`,
          privacyPolicyLink: process.env.NEWSLETTER_PRIVACY_POLICY_URLS?.split(
            ","
          ).find((url) => url.includes(newsletterParticipant.type)),
        }
      );

      newsletterLogger.info(
        `ConfirmNewsletterParticipation newsletterId: ${newsletterId} email: ${newsletterParticipant.email} confirmed`
      );
    } else {
      newsletterLogger.warn(
        `ConfirmNewsletterParticipation: newsletter id: ${newsletterId} email: ${newsletterParticipant.email}  already confirmed`
      );
    }

    return res.status(200).send({ status: "success" });
  } catch (error) {
    newsletterLogger.error(
      "ConfirmNewsletterParticipation err:",
      error as string
    );
    res.status(500).send({ err: "invalid request" });
  }
}

export async function RemoveNewsletterParticipant(req: Request, res: Response) {
  try {
    const { newsletterId } = req.params;

    if (!newsletterId) {
      newsletterLogger.warn(
        "RemoveNewsletterParticipant: invalid request - missing email or type"
      );
      return res.status(400).send({ err: "invalid request" });
    }

    const newsletterParticipant = await Newsletter.findOne({
      where: {
        newsletter_id: newsletterId,
      },
    });

    if (!newsletterParticipant) {
      newsletterLogger.warn(
        `RemoveNewsletterParticipant: newsletter id: ${newsletterId} not found`
      );
      return res.status(400).send({ err: "invalid request" });
    }

    await newsletterParticipant.destroy();

    newsletterLogger.info(
      `RemoveNewsletterParticipant newsletterId: ${newsletterId} removed. Email: ${newsletterParticipant.email}, type: ${newsletterParticipant.type}, state: ${newsletterParticipant.state}`
    );

    return res.status(200).send({ status: "success" });
  } catch (error) {
    newsletterLogger.error("RemoveNewsletterParticipant err:", error as string);
    res.status(500).send({ err: "invalid request" });
  }
}
